@@ -11,29 +11,31 @@
 /**
  * Error codes.
  */
-#define RAFT_NOMEM 1            /* Out of memory */
-#define RAFT_BADID 2            /* Server ID is not valid */
-#define RAFT_DUPLICATEID 3      /* Server ID already in use */
-#define RAFT_DUPLICATEADDRESS 4 /* Server address already in use */
-#define RAFT_BADROLE 5          /* Server role is not valid */
-#define RAFT_MALFORMED 6
-#define RAFT_NOTLEADER 7
-#define RAFT_LEADERSHIPLOST 8
-#define RAFT_SHUTDOWN 9
-#define RAFT_CANTBOOTSTRAP 10
-#define RAFT_CANTCHANGE 11
-#define RAFT_CORRUPT 12
-#define RAFT_CANCELED 13
-#define RAFT_NAMETOOLONG 14
-#define RAFT_TOOBIG 15
-#define RAFT_NOCONNECTION 16
-#define RAFT_BUSY 17
-#define RAFT_IOERR 18        /* File system or storage error */
-#define RAFT_NOTFOUND 19     /* Resource not found */
-#define RAFT_INVALID 20      /* Invalid parameter */
-#define RAFT_UNAUTHORIZED 21 /* No access to a resource */
-#define RAFT_NOSPACE 22      /* Not enough space on disk */
-#define RAFT_TOOMANY 23      /* Some system or raft limit was hit */
+enum {
+    RAFT_NOMEM = 1,        /* Out of memory */
+    RAFT_BADID,            /* Server ID is not valid */
+    RAFT_DUPLICATEID,      /* Server ID already in use */
+    RAFT_DUPLICATEADDRESS, /* Server address already in use */
+    RAFT_BADROLE,          /* Server role is not valid */
+    RAFT_MALFORMED,
+    RAFT_NOTLEADER,
+    RAFT_LEADERSHIPLOST,
+    RAFT_SHUTDOWN,
+    RAFT_CANTBOOTSTRAP,
+    RAFT_CANTCHANGE,
+    RAFT_CORRUPT,
+    RAFT_CANCELED,
+    RAFT_NAMETOOLONG,
+    RAFT_TOOBIG,
+    RAFT_NOCONNECTION,
+    RAFT_BUSY,
+    RAFT_IOERR,        /* File system or storage error */
+    RAFT_NOTFOUND,     /* Resource not found */
+    RAFT_INVALID,      /* Invalid parameter */
+    RAFT_UNAUTHORIZED, /* No access to a resource */
+    RAFT_NOSPACE,      /* Not enough space on disk */
+    RAFT_TOOMANY       /* Some system or raft limit was hit */
+};
 
 /**
  * Size of human-readable error message buffers.
@@ -71,13 +73,24 @@ struct raft_buffer
     size_t len; /* Length of the buffer. */
 };
 
+/**
+ * A type for storing unknown bools.
+ */
+typedef enum {
+    raft_tribool_unknown,
+    raft_tribool_true,
+    raft_tribool_false,
+} raft_tribool;
+#define TO_RAFT_TRIBOOL(b)  ((b) ? raft_tribool_true : raft_tribool_false)
+
 /**
  * Server role codes.
  */
-
-#define RAFT_STANDBY 0 /* Replicate log, does not participate in quorum. */
-#define RAFT_VOTER 1   /* Replicate log, does participate in quorum. */
-#define RAFT_SPARE 2   /* Does not replicate log, or participate in quorum. */
+enum {
+    RAFT_STANDBY, /* Replicate log, does not participate in quorum. */
+    RAFT_VOTER,   /* Replicate log, does participate in quorum. */
+    RAFT_SPARE    /* Does not replicate log, or participate in quorum. */
+};
 
 /**
  * Hold information about a single server in the cluster configuration.
@@ -252,8 +265,9 @@ struct raft_request_vote
  */
 struct raft_request_vote_result
 {
-    raft_term term;    /* Receiver's current term (candidate updates itself). */
-    bool vote_granted; /* True means candidate received vote. */
+    raft_term term;        /* Receiver's current term (candidate updates itself). */
+    bool vote_granted;     /* True means candidate received vote. */
+    raft_tribool pre_vote; /* The response to a pre-vote RequestVote or not. */
 };
 
 /**
@@ -406,6 +420,20 @@ struct raft_io_snapshot_get
     raft_io_snapshot_get_cb cb; /* Request callback */
 };
 
+/**
+ * Asynchronous work request.
+ */
+struct raft_io_async_work;
+typedef int (*raft_io_async_work_fn)(struct raft_io_async_work *req);
+typedef void (*raft_io_async_work_cb)(struct raft_io_async_work *req,
+                                      int status);
+struct raft_io_async_work
+{
+    void *data;                 /* User data */
+    raft_io_async_work_fn work; /* Function to run async from the main loop */
+    raft_io_async_work_cb cb;   /* Request callback */
+};
+
 /**
  * Customizable tracer, for debugging purposes.
  */
@@ -416,6 +444,11 @@ struct raft_tracer
      */
     void *impl;
 
+    /**
+     * Whether this tracer should emit messages.
+     */
+    bool enabled;
+
     /**
      * Emit the given trace message, possibly decorating it with the provided
      * metadata.
@@ -442,7 +475,7 @@ typedef void (*raft_io_close_cb)(struct raft_io *io);
 
 struct raft_io
 {
-    int version;
+    int version; /* 1 or 2 */
     void *data;
     void *impl;
     char errmsg[RAFT_ERRMSG_BUF_SIZE];
@@ -483,11 +516,38 @@ struct raft_io
                         raft_io_snapshot_get_cb cb);
     raft_time (*time)(struct raft_io *io);
     int (*random)(struct raft_io *io, int min, int max);
+    /* Field(s) below added since version 2. */
+    int (*async_work)(struct raft_io *io,
+                      struct raft_io_async_work *req,
+                      raft_io_async_work_cb cb);
 };
 
+/*
+ * version 2:
+ * introduces `snapshot_finalize`, when this method is not NULL, it will
+ * always run after a successful call to `snapshot`, whether the snapshot has
+ * been successfully written to disk or not. If it is set, raft will
+ * assume no ownership of any of the `raft_buffer`s and the responsibility to
+ * clean up lies with the user of raft.
+ * `snapshot_finalize` can be used to e.g. release a lock that was taken during
+ * a call to `snapshot`. Until `snapshot_finalize` is called, raft can access
+ * the data contained in the `raft_buffer`s.
+ *
+ * version 3:
+ * Adds support for async snapshots through the `snapshot_async` function.
+ * When this method is provided, raft will call `snapshot` in the main loop,
+ * and when successful, will call `snapshot_async` by means of the `io->async_work` method.
+ * The callback to `io->async_work` will in turn call `snapshot_finalize`
+ * in the main loop when the work has completed independent of the return value
+ * of `snapshot_async`.
+ * An implementation that does not use asynchronous snapshots MUST set
+ * `snapshot_async` to NULL.
+ * All memory allocated by the snapshot routines MUST be freed by the snapshot
+ * routines themselves.
+ */
 struct raft_fsm
 {
-    int version;
+    int version; /* 1, 2 or 3 */
     void *data;
     int (*apply)(struct raft_fsm *fsm,
                  const struct raft_buffer *buf,
@@ -496,6 +556,14 @@ struct raft_fsm
                     struct raft_buffer *bufs[],
                     unsigned *n_bufs);
     int (*restore)(struct raft_fsm *fsm, struct raft_buffer *buf);
+    /* Fields below added since version 2. */
+    int (*snapshot_finalize)(struct raft_fsm *fsm,
+                             struct raft_buffer *bufs[],
+                             unsigned *n_bufs);
+    /* Fields below added since version 3. */
+    int (*snapshot_async)(struct raft_fsm *fsm,
+                          struct raft_buffer *bufs[],
+                          unsigned *n_bufs);
 };
 
 /**
@@ -1011,6 +1079,19 @@ RAFT_API void raft_heap_set(struct raft_heap *heap);
  */
 RAFT_API void raft_heap_set_default(void);
 
+/**
+ * Return a reference to the current dynamic memory allocator.
+ *
+ * This is intended for use by applications that want to temporarily replace
+ * and then restore the original allocator, or that want to defer to the
+ * original allocator in some circumstances.
+ *
+ * The behavior of attempting to mutate the default allocator through the
+ * pointer returned by this function, including attempting to deallocate
+ * the backing memory, is undefined.
+ */
+RAFT_API const struct raft_heap *raft_heap_get(void);
+
 #undef RAFT__REQUEST
 
 #endif /* RAFT_H */
